// -*- C++ -*-

#include "bpgraph.hpp"

// #include <ejg/log.hpp>

#include <boost/graph/connected_components.hpp>
#include <boost/graph/properties.hpp>
#include <boost/pending/disjoint_sets.hpp>
#include <boost/graph/incremental_components.hpp>

#include <sstream>
#include <cassert>
#include <limits>
#include <cmath>

using namespace cmb;
using namespace std;
using namespace boost;

namespace cmb
{
  template <typename T>
  ostream& operator <<(ostream& out, const vector<T>& x)
  {
    for (typename vector<T>::const_iterator i = x.begin();
         i != x.end();
         ++i)
    {
      out << (i == x.begin() ? "<" : ", ") << *i;
    }
    return out << ">";
  }

  namespace bp
  {
    typedef boost::graph_traits<bp::Graph>::vertex_descriptor VertDesc; // uint
    typedef boost::graph_traits<bp::Graph>::edge_descriptor EdgeDesc;   // uint
    typedef boost::graph_traits<bp::Graph>::vertex_iterator VertIter;   
    typedef boost::graph_traits<bp::Graph>::edge_iterator EdgeIter;    
  }
}

static const unsigned BAD = numeric_limits<unsigned>::max();

cmb::bp::Vert::Vert() : label(0), positive(true), cycle(BAD) {}
cmb::bp::Edge::Edge() : reality(true), cycle(BAD) {}

cmb::cg::Vert::Vert() : min(BAD), max(BAD), directed(false), component(BAD) {}
cmb::cg::Edge::Edge() : component(BAD) {}

static bool odd(int x)
{
  return x & 1;
}

static bool even(int x)
{
  return !odd(x);
}

// canonical index: 0 1 2 3 ... n-1 n
// canonical gene:  0 1 2 3 ... n-1 n
// source gene:     0 a b c ...  y  z  (set[a-z] == set[1-n])

// graph size = 2n+2
// graph index:  0  1    2  3  ... 2n 2n+1
// graph label: -0  0+  -1  1+ ... -n   n+

// a linear graph will have a zero added at the front
// otherwise it is the same as a cyclic graph; including
cmb::bp::Graph cmb::bp::make(const Seq& x, GraphType t)
{
  bool linear = t == LINEAR;
  Seq src(x);
  if (linear) src.insert(src.begin(), 0); // marks endpoints of line

  int n = src.size();

  // prepare an inverse map
  std::vector<unsigned> inv(n + 1, BAD); // waste one in linear case
  for (int i = 0; i < n; ++i)
    {
      assert(abs(src[i]) < n);
      assert(inv[abs(src[i])] == BAD);
      inv[abs(src[i])] = i;
    }

  Graph g(2*n);
  for (int k = 0; k < 2*n; ++k)
    {
      int gene = src[k/2];
      
      g[k].label = abs(gene);
      g[k].cycle = BAD;
      
      // construct reality edges, odd to even
      if (odd(k))
        {
          g[k].positive = (gene >= 0);
          pair<EdgeDesc,bool> e = add_edge(k, (k+1)%(2*n), g); // reality edge
          assert(e.second);
          g[e.first].reality = true;
          g[e.first].cycle = BAD;
        }
      else 
        {
          g[k].positive = (gene < 0);
        }

      if (g[k].positive) // desire edges connect + to - (note cycle)
        {
          int i = inv[(abs(gene) + 1) % n];
          int v = 2 * i + (src[i] < 0 ? 1 : 0);
          pair<EdgeDesc,bool> e = add_edge(k, v, g);
          assert(e.second);
          g[e.first].reality = false;
          g[e.first].cycle = BAD;
        }
    }
  return g;
}

struct Count
{
  unsigned cycle;
  unsigned count;
};

typedef std::vector<Count> Stack;

cmb::cg::Graph cmb::cg::make(bp::Graph& bp)
{
  // find all cycles == connected components in break point graph
  unsigned m = connected_components(bp, get(&bp::Vert::cycle, bp));

  Graph cg(m); // cycle graph: each cycle is a vertex, each crossing an edge

  // discover directed cycles (cycles with a backwards desire edge)
  for (Range<bp::EdgeIter> i = edges(bp); i; ++i)
    {
      if (!bp[*i].reality && even(source(*i, bp)) == even(target(*i, bp)))
        {
          cg[bp[*i].cycle].directed = true;
        }
    }

#ifdef NEWCODE

  // find components.  If two cycles cross, then they are in the same
  // component.

  // let bp be the break point graph and cg the cycle graph
  // each cycle in bp is a vertex in cg
  // if cycles cross in bp, then each is reachable from the other in cg and
  // equivlently are both part of the same component in cg.  Note that
  // this does not require that all crossings in bp map to edges in cg.
  // That is important, because there may be as many as O(n^2) crossings.

  // Naive Algorithm:
  // map each bp::cycle to/from cg::vertex
  //
  // useing a bp::edge stack, iterate over bp::graph vertices, lowest - highest
  //   get desire edge "d"
  //   if d is forward (ie: other end is higher indexed vertex)
  //     push d
  //   else
  //     find d in stack
  //     connect cg(d) to cg(x) for all x between stack top and d
  //     remove d from stack
  // find connected components in cg::graph

  // this should work, but is O(n^2)

  // Seemingly simpler algorithm:
  // same thing, but use component on stack instead of arc
  // stack can be compressed: component c on stack n times cccc => (c, 4)
  // now not O(n^2) in stack searches, but...
  // must keep track of what component each edge is in
  // this task becomes O(n^2), since it may change many times

  // Implemented algorithm:
  // use union-find to keep track of edge component membership
  // this leads to O(n\alpha(n)) where \alpha is the inverse ackerman function

  std::vector<unsigned> parent;
  disjoint_sets<unsigned, unsigned*> ds(get(&cg::Vert::component, cg), 
                                        &parent[0]);
  initialize_incremental_components(cg, ds);
  // no edges yet, so below not needed
  // incremental_components(cg, ds);
  
  for (Range<bp::VertIter> i = vertices(bp); i; ++i)
    {
      Range<bp::EdgeIter> j = out_edges(*i, bp);
      EdgeDesc d = (bp[*j].reality ? *++j : *j++); // desire edge
      assert(j && !++j); // 1 reality and 1 desired edge per vertex
      assert(!bp[d].reality);

      unsigned target = target(d, bp);
      if (target == *i)
        {
          target = source(d, bp);
        }
      else
        {
          assert(target(d, bp) == *i);
        }
      assert(target != *i);
      assert(bp[*i].cycle == bp[target].cycle);
      unsigned component = ds.find_set(bp[target].cycle);
      
      if (target > *i) // forward edge
        {
          if (stack.back().component == component)
            {
              ++stack.back().count;
            }
          else
            {
              stack.push_back(new Count(component, 1));
            }
        }
      else        // backward edge
        {
          assert(!stack.empty());
          while (ds.find_set(stack.back().component) != ds.find_set(component))
            {
              Count x = stack.back();
              stack.pop_back();
              assert(!stack.empty());

              stack.back().count += x.count;
              ds.union_set(stack.back().component, x.component);
              pair<EdgeDesc,bool> e = add_edge(stack.back().component, 
                                               x.component, cg);
              assert(e.second);
            }
          assert(stack.back().count > 0);
          --stack.back().count;
          if (stack.back().count == 0)
            {
              stack.pop_back();
            }
        }
    }

#if 0
  Components x;
  for (Range<VertIter> i = vertices(cg); i; ++i)
    {
      while (x.size() <= cg[*i].component)
        {
          x.push_back(Exemplar(BAD, false));
        }
      if (cg[*i].directed || x[cg[*i].component] == BAD)
        {
          x[cg[*i].component].exemplar = *i;
          x[cg[*i].componnet].directed = cg[*i].directed;
        }
    }
#endif
#endif

  return cg;
}

// ...[node][gap][label][gap][node][arc][node][gap][label]...

static const double      X0 =  0.0; // left of graph
static const double      Y0 =  0.0; // bottom of graph
static const double  LABELX = 14.0; // one character (includes buffer gap)
static const double  LABELY =  8.0; // one character
static const double   NODEX =  3.0; // radius of a node (graph vertex)
static const double    ARCX =  5.0; // edge to edge node to node space
static const double  BARGAP =  3.0; // gap from top of char to bar
static const double BARFRAC =  0.5; // fraction of LABELX for bar
static const double MARGINY =  2.0; // svg drawing margin above and below
static const double MARGINX =  2.0; // svg drawing margin right and left
static const double ARCFRAC =  0.85; // height reduction for arcs

static const double BLOCKX = LABELX + ARCX + 2 * NODEX + ARCX;
static const double MIDY   = Y0 - LABELY / 2.0;
static const double BARY   = Y0 - (LABELY + BARGAP);

// 0 o--o 1 o--o 2 o--o N
static double nodex(unsigned i)
{
  double x = X0 + (i/2) * BLOCKX + (odd(i) ?
                                    (LABELX + NODEX / 2) :
                                    (- NODEX / 2));
  return x;
}

static double nodex(unsigned i, unsigned n)
{
  return (i ? nodex(i) : nodex(n));
}

// 0 o--o 1 o--o 2 o--o N
static double labelx(unsigned i)
{
  return X0 + LABELX / 2.0 + i * BLOCKX;
}

static void svg_dot(ostream& out, const std::string& tab,
                    double x, double y, double radius)
{
  out << tab << "<circle fill='#000000' stroke='none' "
      << "cx='" << x << "' cy='" << y << "' r='" << radius << "'/>\n";     
}

static void svg_ctext(ostream& out, const std::string& tab,
                      double x, double y, const std::string& text)
{
  out << tab << "<text x='" << x << "' y='" << y << "'>" 
      << text << "</text>\n";
}

static void svg_cbox(ostream& out, const std::string& tab,
                     double x, double y, double width, double height)
{
  out << tab << "<rect fill='none' stroke='#000000' "
      << "x='" << x << "' y='" << y << "' "
      << "width='" << LABELX << "' height='" << LABELY << "'/>\n";
}

static void svg_hline(ostream& out, const std::string& tab,
                      double x0, double x1, double y)
{
  out << tab << "<path fill='none' stroke='#000000' "
      << "d='M " << x0 << "," << y << " "
      << "L " << x1 << "," << y << "'/>\n";

}

static void svg_harc(ostream& out, const std::string& tab,
                     double x0, double x1, double y)
{
  double r = abs(x1 - x0) / 2.0;
  out << tab << "<path fill='none' stroke='#000000' "
      << "transform='scale(1.0, " << ARCFRAC << ")' " 
      << "d='M " << x0 << "," << y << " "
      << "A " << r << "," << r 
      << " 0 0 1 " << x1 << "," << y << "'/>\n";            
}

static void svg_lbl(ostream& out, const std::string& tab,
                    double x, double y, unsigned label, bool negative,
                    double width, double height)
{
  ostringstream s;

  if (label)
    {
      s << label;
      if (negative)
        {
          svg_hline(out, tab, x - width / 2, x + width / 2, -height);
        }
    }
  else if (negative)
    {
      s << "R";
    }
  else
    {
      s << "L";
    }
  svg_ctext(out, tab, x, y, s.str());
}

#if 0
static void svg_page(ostream& out, const std::string& tab,
                     double x, double y,
                     double width, double height, 
                     const std::string& color)
{
  out << tab << "  <rect fill='#" << color << "' stroke='none' "
      << "x='" << x << "' y='" << y 
      << "' width='" << width << "' height='" << height << "'/>\n";
}
#endif

static void svg_nodes(ostream& out, const std::string& tab, bp::Graph& g)
{
  unsigned n = num_vertices(g);

  for (unsigned i = 0; i < n; ++i)
    svg_dot(out, tab, nodex(i, n), MIDY, NODEX / 2.0);

  for (unsigned i = 0; i < n/2; ++i)
    svg_lbl(out, tab, 
            labelx(i), Y0, g[2*i].label, g[2*i].positive,
            LABELX * BARFRAC, LABELY + BARGAP);
  if (n > 0)
    {
      svg_lbl(out, tab, labelx(n/2), Y0, 0, true, 0, 0);
    }

  if (false)
    svg_cbox(out, tab, labelx(0), Y0 + LABELY, LABELX, LABELY);
}

static std::pair<double, double> endpoints(bp::EdgeDesc e, const bp::Graph& g)
{
  unsigned s = source(e, g);
  unsigned t = target(e, g);


  if (s == 0) s = num_vertices(g);
  if (t == 0) t = num_vertices(g);

  std::pair<double, double> result(nodex(s), nodex(t));
  if (result.first > result.second) swap(result.first, result.second);

  assert(result.first < result.second); // ie !=
  
  return result;
}

static void svg_edges(ostream& out, const std::string& tab, bp::Graph& g)
{
  for (Range<bp::EdgeIter> i = edges(g); i; ++i)
    {
      std::pair<double, double> v = endpoints(*i, g);
      if (g[*i].reality)
        {
          svg_hline(out, tab, v.first, v.second, MIDY); // reality edge
        }
      else
        {
          svg_harc(out, tab, v.first, v.second, MIDY);  // desire edge
        }
    }
}

static double svg_height(bp::Graph& g)
{
  double h = 0.0;
  for (Range<bp::EdgeIter> i = edges(g); i; ++i)
    {
      std::pair<double, double> v = endpoints(*i, g);
      h = std::max(h, abs(v.second - v.first)/2);
    }
  return ARCFRAC * h + LABELY / 2;
}

static double svg_width(bp::Graph& g)
{
  unsigned n = num_vertices(g);
  return nodex(n) + LABELX;
}

void cmb::bp::drawSVG(ostream& out, bp::Graph& g)
{
  double SCALE = 1.0;

  double width = svg_width(g);
  double height = svg_height(g);
  double page_width = SCALE * (width + 2 * MARGINX);
  double page_height = SCALE * (height + 2 * MARGINY);

  out << "<?xml version='1.0' encoding='UTF-8' standalone='no'?>\n"
//       << "<!DOCTYPE svg PUBLIC '-//W3C//DTD SVG 1.1//EN'\n"
//       << "  'http:://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd'>\n"
      << "<!-- break point graph -->\n"
      << "<svg version='1.1' baseProfile='full'\n"
      << " width='" << (page_width) << "' height='" << (page_height) << "'\n"
      << " xmlns='http://www.w3.org/2000/svg'\n"
      << " xmlns:xlink='http://www.w3.org/1999/xlink'\n"
      << " xmlns:ev='http://www.w3.org/2001/xml-events'>\n";
  // svg_page(out, "  ", 0, 0, page_width, page_height, "AAAAFF");
  out << "  <g transform='scale(" << SCALE << "," << SCALE << ")'>\n"
      << "    <g font-size='12px' text-anchor='middle' font-family='Bitstream Vera Sans' "
      << "transform='translate(" << MARGINX 
      << ", " << (height+MARGINY) << ")'>\n";

  string tab = "      ";

  // svg_page(out, tab, 0, 0, width, height, "AAAAAA");

  svg_nodes(out, tab, g);
  svg_edges(out, tab, g);
  
  out << "    </g>\n";
  out << "  </g>\n";
  out << "</svg>\n";
}

void cmb::bp::drawSVG(ostream& out, const Seq& s, GraphType t)
{
  Graph g = make(s,t);
  drawSVG(out, g);
}

