// Krister Swenson, Mark Marron                               Spring 2003
//
//The functions used to find the minimum edit distance from the identity.
//
//Explanation about insMatrix:
//   The number of locations that a given insertion can occur can grow through
//   reversals.  We keep track of all the places the insertions can occur in
//   the -insMatrix- array.  The identity of the insertions are anonymous but
//   could easily be recovered.
//

#include "med.h"

// #define VERBOSE

using namespace std; 
using namespace __gnu_cxx; 

/* Return a Minimum Edit Distance estimate from the identity permutation to the
given sequence -seq-.  The operations permitted are reversals, insertions, and
deletions. Copies should be handled outside of this function.
  -targlen- is the length of the target sequence.
  -largest- all values above this in -seq- are inserted elements.
            (the length of the source sequence)
*/
int getMED(int* seq, int targlen, int largest, int &inv, int &del, int &ins)
  {
  int i;

  //dupCheck(seq, targlen);
                                  //The number of inserts (not in the identity).
  int numins = getNumInserts(seq, targlen, largest);

  stack<Pair> dels;               //Keeps track of the deletions done.
                                  //Fill -dels- and rename -seq- to contain
                                  //  ONLY CONTIGUOUS ELEMENTS.
  findDels(dels, seq, targlen, largest);
                                  //The length without the deletions.
  int lenwod = largest - dels.size();

  //cerr << "+++++++++++++++++++++++++++++++lenwod" << lenwod << endl;
  //cerr << "                      +++++++++targlen" << targlen << endl;
  //cerr << "                      +++++++++numins" << numins << endl;
  //cerr << "                      +++++++++dels" << dels.size() << endl;
  int numpos = lenwod + 1;        //The number of spots insertions could be.

  int** insMatrix 
    = new int*[numpos];		  //A spot for each insertion for each position
  for(i=0; i < numpos; i++)	  //Initialize insMatrix:
    {
    insMatrix[i] = new int[numins];
    for(int j=0; j < numins; j++)
      insMatrix[i][j] = 0;
    }                             //Fill insMatrix with the appropriate values
                                  //and remove the inserts.
  //cerr << "Confirm and fill" << endl;                                
  //printArray(seq, lenwod+numins);
  //confirmRenaming(seq, lenwod+numins);
  //confirmRenaming(seq, lenwod);
  fillInsMatrix(insMatrix, seq, lenwod, numpos, numins, largest-dels.size());

#ifdef SORTING_SEQUENCE
  // Save info on genome 2 sequence after removing insertions
  if (globalSortSequence) {
    globalSortSequence->setConsensusG2(seq, lenwod);
  }
#endif //SORTING_SEQUENCE

  //printArray(seq, lenwod);
  //confirmRenaming(seq, lenwod);

  vector<Pair> pi;		    //Keeps track of the reversals done.
  Ograph gr(seq, lenwod);           //Build the graph.
  int deletions = gr.markDels(dels);//Mark edges that have deletions (El-M).

  int reversals = gr.resolve(pi);   //The number of reversals.

  //-=-=-=-=-=-=-=-=-=-=-=-=-=-= verification: -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  /*
  vector<Pair> opi(pi.rbegin(), pi.rend());
  int* oseq = gr.getSeq();

  cerr << endl ;
  for(i=0; i < reversals; i++)
    {
    Pair inter = opi.back();
    int min = inter.getLeft();
    int max = inter.getRight();
    while(min <= max)
      {
      int t;
      t = oseq[min];
      oseq[min] = -oseq[max];
      oseq[max] = -t;

      min++;
      max--;
      }
    opi.pop_back();
    }

  //for(i=0; i<lenwod; i++)
    //cerr << seq[i] << " " << oseq[i] << endl;

  for(i=0; i<lenwod; i++)
    if(seq[i] != oseq[i])
      cerr << "MISMATCH:" << seq[i] << " " << oseq[i] << endl;

  free(oseq);
  */
  //-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-

#ifdef OVERLY_OPTIMISTIC_INSERTION_IDEA
  expandInsMatrix(insMatrix, pi, reversals, numpos, numins);
#endif // OVERLY_OPTIMISTIC_INSERTION_IDEA
  //findClumps(insMatrix, numpos, numins);
  int insertions;
  insLocs(insMatrix, numpos, numins, &insertions);
  killInsMatrix(insMatrix, numpos);
#ifdef SORTING_SEQUENCE
  if (globalSortSequence) {
//     globalSortSequence->printGenomes();
//     globalSortSequence->printExecutedReversals();
#ifdef OVERLY_OPTIMISTIC_INSERTION_IDEA
    globalSortSequence->computeInsConfigurations();
#else // OVERLY_OPTIMISTIC_INSERTION_IDEA
    globalSortSequence->computeStickyInsMap();
#endif // OVERLY_OPTIMISTIC_INSERTION_IDEA
    globalSortSequence->computeDelPositions();
//     globalSortSequence->printOriginalHurdles();
//     vector<int> midVector;
//     globalSortSequence->getMidpointGenome(midVector);
//     cout << "MIDPOINT GENOME" << endl;
//     for (vector<int>::iterator midIt = midVector.begin();
// 	 midIt != midVector.end(); ++midIt)
//       cout << *midIt << " ";
//     cout << endl;
  }
#endif // SORTING_SEQUENCE

#ifdef VERBOSE
  cout << "Length_of_S1: " << largest << endl;
  cout << "Length_of_S2: " << targlen << endl;
  cout << "Insertions: " << insertions << endl;
  cout << "Reversals: " << reversals << endl;
  cout << "Deletions: " << deletions << endl;
#endif // VERBOSE

  inv = reversals; del = deletions; ins = insertions;

  return reversals+deletions+insertions;
  }


int getMED(int* targ, int targlen, int largest) {
  int inv, del, ins;
  return getMED(targ, targlen, largest, inv, del, ins);
}





//----------------------------------------------------------------------------


/* Return the number of values that were inserted from a source permutation
with -srclen- elements to give -seq-.
*/
int getNumInserts(int seq[], int origlen, int srclen)
  {
  int count = 0;
  while(--origlen >= 0)
    if(abs(seq[origlen]) > srclen)
      count++;

  return count;
  }


/* Find the deletions (elements missing in seq that would be in the identity
1 2 ... srcsize).  Rename the sequence to be contiguous.
*/
void findDels(stack<Pair> &dels, int seq[], int seqsize, int srcsize)
  {
  int i;
  int maxel = 0;		//The maximum element value in the source or
				//target.
  Pair currDelete;
  maxel = getMax(seq, seqsize);
  if (maxel < srcsize)
    maxel = srcsize;

  int** survivors;
  survivors = (int**) calloc(maxel, sizeof(int*));
				//Pointers to each element of final sequence.
  for(i=0; i < maxel; i++)	// Used for renaming the sequence.
    survivors[i] = NULL;
  

  for(i=0; i < seqsize; i++)		//For each value in the sequence
    survivors[abs(seq[i])-1] = &seq[i];	//set a pointer.

  int spot = 1;
  int last = 0;			//The last node seen.
  for(i=0; i < maxel; i++)	//Rename the non-inserts using -survivors-:
    {
    if(survivors[i])		//If there is a value -i- in the sequence
      {
      //cerr << "  " << *survivors[i] << " ---> ";
#ifdef SORTING_SEQUENCE
      if (globalSortSequence) {
	globalSortSequence->mapConsensusToRenamed(spot, abs(*survivors[i]));
      }
#endif //SORTING_SEQUENCE
      if(*survivors[i] > 0)	//rename it
        *survivors[i] = spot++;
      else
        *survivors[i] = -(spot++);
      //cerr << "  " << *survivors[i] << endl;
      last = *survivors[i];
      }
    else if(i < srcsize)
      {
      //cerr << "DEL: " << i+1 << endl;
      currDelete = Pair(i+1,last);
      dels.push(currDelete);	//This deletion should be part of the last 
				//  element's label. (missing, last seen).
#ifdef SORTING_SEQUENCE
      if (globalSortSequence) {
	globalSortSequence->addDelete(currDelete);
      }
#endif //SORTING_SEQUENCE
      }
    }
  free(survivors);
  //cerr << endl << "------------------------------------------" << endl;
  //cerr << "print array" << endl;
  //printArray(seq, seqsize);
  //cerr << "confirm renaming" << endl;
  //confirmRenaming(seq, seqsize);
  //cerr << endl << "------------------------------------------" << endl;
  }



/* Return the maximum element.
*/
int getMax(int a[], int asize)
  {
  int maxel = 0;
  for(int i=0; i < asize; i++)
    if(maxel < abs(a[i]))
{
//cerr << "----     - --" << a[i-1] <<   "- -----" << a[i] << "--" << i 
//     << endl;
      maxel = abs(a[i]);
}

  return maxel;
  }



/* Find the clump sizes of insertions. This is obsolete.
*/
// void findClumps(int** insMatrix, int numpos, int numins)
//   {
//   int ins, pos;
//   int clump;                        //The current clump length.
//   int* largest;
//   largest = (int*) calloc(numins, sizeof(int));
//					//Position of largest clump for a given
//					//  insertion.
//   bool* alone;
//   alone = (bool*) calloc(numins, sizeof(bool));
//					//True if not assigned to a clump yet.

//   for(ins=0; ins < numins; ins++)	//Call the first row the largest clump
//     {
//     largest[ins] = 0;		//  for a given insertion initially.
//     alone[ins] = true;
//     }

//   for(pos=0; pos < numpos; pos++)	//Compile clump sizes along positions:
//     {
// /*
// print2D(insMatrix, numpos, numins);
// cerr << endl;
// */
//     clump=0;
//     for(ins=0; ins < numins; ins++)
//       {
//       if(insMatrix[pos][ins])	//Update the clump value.
//         {
//         clump++;
//         insMatrix[pos][ins] = clump;
//         }

//       if(insMatrix[largest[ins]][ins] < clump)
//         largest[ins] = pos;		//Update the largest array.
//       }
//     }

// /*
// cerr << endl;
// print2D(insMatrix, numpos, numins);
// printArray(largest, numins);
// */
//					//Pick the clumps greedily:
//					//  **Efficiency(stk)
//   for(ins = numins-1; ins >= 0; ins--)
//     if(alone[ins])
//       {
//   //    fprintf(stderr, "\nINS: position-%d indeces-%d", largest[ins], ins);

//       bool notdone = true;
//       for(int i=ins-1; i >= 0 && notdone; i--)
//         if(insMatrix[largest[ins]][i])
//           {
//           if(insMatrix[largest[ins]][i] == 1)
//             notdone = false;
//     //      cerr << " " << i;
//           alone[i] = false;
//           }
//       }
//     free(largest);
//     free(alone);
//   //cerr << endl;
//   }



/*Fill the insMatrix array according to the extra values from the sequence.
*/
void fillInsMatrix(int** insMatrix, int seq[], int seqlen, int numpos,
               int numins, int srclen)
  {
  //cerr << "\nFill insMatrix: numins=" << numins << " numpos=" << numpos 
  //     << endl;
  int pushfactor = 0;		       //Number of insertions detected so far.
  for(int i=0; i < numins+seqlen; i++) //Remove the inserts from the string:
    {
    seq[i - pushfactor] = seq[i];      //Compact the string.
    if(abs(seq[i]) > seqlen)	       //If an insertion is found
      {
      //cerr << "\tGotit:" << abs(seq[i]) << " > " << seqlen << " || col:"
      //     << abs(seq[i]) - seqlen - 1 << "  row:" << i-pushfactor << endl;
      //insMatrix[i - pushfactor][abs(seq[i]) - seqlen - 1] = 1;
      //cerr << "\tGotit:" << abs(seq[i]) << " > " << seqlen << " || col:"
      //     << pushfactor << "  row:" << i-pushfactor << endl;
      insMatrix[i - pushfactor][pushfactor] = 1;
      pushfactor++;		       //mark it as such.
#ifdef SORTING_SEQUENCE
      if (globalSortSequence) {
	globalSortSequence->addInsert(Pair(seq[i],i));
      }
#endif //SORTING_SEQUENCE
      }
    }
  }


/*Free the memory from insMatrix.
*/
void killInsMatrix(int** insMatrix, int numpos)
  {
  for(int i=0; i < numpos; i++)
    delete [] insMatrix[i];
  delete [] insMatrix;
  }

/*Confirm that there is at least one 1 per column
      (an insertion must be somewhere).
*/
void confirmInsMatrix(int** insMatrix, int numpos, int numins)
  {
  for(int i=0; i < numins; i++)
    {
    bool found = false;
    for(int j=0; j < numpos; j++)
      if(insMatrix[j][i])
        found = true;

    if(!found)
      cerr << "Error[confirmInsMatrix]: Column " << i << " missing a 1!" 
	   << endl;
    }
  }


/*Fill insMatrix with the possible insertions through -pi-.
*/
void expandInsMatrix(int** insMatrix, vector<Pair> &pi, int stepCount,
                    int numpos, int numins)
  {
  int left, right;
  int i, j;
  int *temp;

  // step thru pi backwards so the reversals go from genome 2 (which contains
  // the insertions) to genome 1
  for(i=stepCount-1; i >= 0; i--)
    {
    // get left and right of current reversal
    left = pi[i].getLeft();
    right = pi[i].getRight()+1;

    //cerr << "reverse (" << left << "," << right << ")" << endl;

    // merge insertions for the genes on the two sides of the reversal
    for(j=0; j < numins; j++)
      {
      insMatrix[left][j] = insMatrix[right][j] 
	= insMatrix[left][j] | insMatrix[right][j];
      }

    // reverse the consensus genes (insMatrix rows) within the reversal
    ++left;
    --right;
    while(left < right)
      {
      temp = insMatrix[left];
      insMatrix[left] = insMatrix[right];
      insMatrix[right] = temp;
      ++left;
      --right;
      }
    }
#ifdef SORTING_SEQUENCE
  if (globalSortSequence) {
    globalSortSequence->setInsMatrix(insMatrix, numpos, numins);
    //globalSortSequence->printInsMatrix();
  }
#endif // SORTING_SEQUENCE
  //print2D(insMatrix, numpos, numins);
  }







/* Print an integer array.
*/
void printArray(const int a[], int seqsize)
  {
  int i;

  for(i=0; i < seqsize-1; i++)
    fprintf(stderr, "%d ", a[i]);

  fprintf(stderr, "%d\n", a[seqsize-1]);
  }



/* Print a 2-dimensional integer array.
*/
void print2D(int** a, int numpos, int numins)
  {
  int i,j;

  for(i=0; i < numpos; i++)
    {
    fprintf(stderr, "%.3d- ", i);
    for(j=0; j < numins; j++)
      fprintf(stderr, "%d ", a[i][j]);

    fprintf(stderr, "-%.3d\n", i);
    }
  }


/* Print missing element in the sequence if they exist.
*/
void confirmRenaming(int s[], int seqlen)
  {
  int i;
  cerr << "-- confirming... LENGTH:" << seqlen << endl;
  int* a; 
  a = (int*) calloc(seqlen, sizeof(int));
  for(i=0; i < seqlen; i++)
    a[i] = abs(s[i]);
  bsort(a, seqlen);
  for(i=1; i < seqlen; i++)
    if(a[i]-1 != a[i-1])
      cerr << "==============SKIPPED:" << a[i] << "," << a[i-1] << endl;
  free(a);
  cerr << "done." << endl;
  }

void hasElement(int elem, int s[], int seqlen)
  {
  for(int i=0; i < seqlen; i++)
    if(elem == abs(s[i]))
      cerr << "HAS:" << elem << endl;
  }

int* readInSeq(int &len, int &numCopies, int &srclen)
  {
  int* input;
  int i;

  cin >> srclen;
  cin >> numCopies;
  cin >> len;

  input = new int[len];

  for(i = 0; i < len; ++i)
    cin >> input[i];

  return input;
  }




/* Return true if val is a member of a.
**Efficiency -> use set.
*/
bool isMember(int val, int a[], int len)
  {
  while(len--)
    if(val == a[len])
      return true;

  return false;
  }


/* Print an error and die if there are duplicates in the given sequence.
*/
void dupCheck(const int* seq, const int len)
  {
  hash_map< int, int, hash<int> > beenseen; //True if the value has been seen.

  int i;
  for(i=0; i < len; i++)                    //Mark those elements that exist.
    {
    if(beenseen[abs(seq[i])])
      {
      cerr << "!!Serious Error!! ";
      cerr << "Two " << abs(seq[i]) << "s found at index " << i << " and "
           << beenseen[abs(seq[i])] << "." << endl;
      exit(1);
      }
    beenseen[abs(seq[i])] = i;
    }
  }

