/* Krister Swenson                                                  Spring 2006

Provide functions for dealing with sorting lattices.
*/

#include "latticefuncs.h"

using namespace std; 
using namespace __gnu_cxx; 


  //The integer (hash value) to sequence (string) reverse lookup.
hash_map< unsigned int, string, hash<int> >seqlookup;



/* Find an island of size equal to the inversion distance.
 If one exists return true.
*/
bool hasIsland(Ograph &gr)
  {
  string seqstr;
  seqstr = gr.getSeqString();
  hash_map<unsigned int, Vertex> vertexset;
  Vertex ub(DEKHash(seqstr));             //Make the upper bound,
  vertexset[ub.getLabel()] = ub;          //then add it to the vertex set
  fullLattice(vertexset, ub, gr, seqstr); //and build the lattice from it.

  int* ID = makeID(gr.getSeqLen());
  string IDstr = seqTOstring(ID, gr.getSeqLen());
  free(ID);
  if(hasIsland(vertexset[ub.getLabel()], vertexset[DEKHash(IDstr)]))
    return true;

  return false;
  }


/* Find an island of size equal to the inversion distance. If one exists return
true.
We do this by building a connected component (not including links
through the id) from an upper bound's child.  If after this is done there are
any unvisited children we know we have an island.
*/
bool hasIsland(Vertex &ub, Vertex &lb)
  {
  ub.mark();    //Mark the ub so we don't find a path through it
  lb.mark();    //Mark the lb so we don't find a path through it.

  if(ub.getDegree() < 2)
    return false;

  markComponent(*(ub.nBeg()->second));
  for(int i=1; i<ub.getDegree(); i++) //If any of the other children
  for(vpmap::const_iterator i=ub.nBeg(); i != ub.nEnd(); i++)
    if(!i->second->getMark())         //are not maked,
      {
      clean(*(ub.nBeg()->second));    //unmark the dirtied vertices,
      return true;                    //and indicate we have found one.
      }

  clean(*(ub.nBeg()->second));        //Unmark the dirtied vertices.
  return false;
  }


/* Mark all vertices reachable from -v-.
*/
void markComponent(Vertex &v)
  {
  v.mark();
  for(vpmap::const_iterator i=v.nBeg(); i != v.nEnd(); i++)
    if(!i->second->getMark())
      markComponent(*(i->second)); 
  }





/* Return true if one vertex is reachable from the other.
*/
bool isReachable(Vertex &v, const Vertex &target)
  {
  v.mark();
  for(vpmap::const_iterator j=v.nBeg(); j != v.nEnd(); j++)
    {
    if(j->second == &target)           //If we found the target vertex.
      return true;                     //return true,
    else if(!j->second->getMark())     //otherwise recurse to unmarked vertices.
      isReachable(*(j->second), target);
    }
  return false;
  }

/* Unmark v's connected component induced by considering only marked vertices.
*/
void clean(Vertex &v)
  {
  v.unmark();
  for(vpmap::const_iterator i=v.nBeg(); i != v.nEnd(); i++)
    if(i->second->getMark())         //For each maked neighbor,
      clean(*(i->second));           //unmark it and it's neighbors.
  }


/* Get the sorting lattice.
*/
void fullLattice(hash_map<unsigned int, Vertex> &vset,
                 Vertex &v, Ograph &gr, const string &str)
  {
  vector<Pair> oOps;

  gr.getSafeInvs(oOps);              //Get the safe ops on divergent edges,
  for(int i=0; size_t(i)<oOps.size(); i++)
    {                                //for each divergent op
    Ograph newgr = gr;               //make a new graph
    newgr.revAll(oOps[i]);           //that we do one of the inversions on.
    string newstr;
    newstr = newgr.getSeqString();   //Output the sequence

    Vertex newv = Vertex(DEKHash(newgr.getSeqString()));
    if(!lookup(vset, newv.getLabel())) //If the child is not in the vertex set
      vset[newv.getLabel()] = newv;    // then we add it.
                                       //Make new vertex adjacent to old,
    vset[v.getLabel()].addNeighbor(&vset[newv.getLabel()]);
                                       //make old vertex adjacent to new,
    vset[newv.getLabel()].addNeighbor(&vset[v.getLabel()]);

    if(newgr.notSorted())
      fullLattice(vset, newv, newgr, newstr);//and continue to sort.
    }
  }



/* Given the upper bound, traverse the graph and only print the downward edges
so that dot makes it look like a lattice.
  (i.e. do a breadth-first traversal)
*/
void printLattice(Vertex &ub)
  {
  queue<Vertex*> Q;

  Q.push(&ub);

  while(!Q.empty())
    {
    Vertex* v = Q.front();
    Q.pop();
    v->mark();
    string vstr = seqlookup[v->getLabel()];
    for(vpmap::const_iterator i=v->nBeg(); i != v->nEnd(); i++)
      {
      if(!i->second->getMark())
        {
        cerr << "  \"" << vstr << "\\n\" -> \""
             << seqlookup[i->second->getLabel()] << "\\n\"" << endl;
        Q.push(i->second);
        }
      }
    }
  }


/* Traverse the graph and print the edges.
*/
void printGraph(Vertex &v)
  {
  v.mark();
  string vstr= seqlookup[v.getLabel()];
  for(vpmap::const_iterator i=v.nBeg(); i != v.nEnd(); i++)
    {
    cerr << "  \"" << vstr << "\\n\" -> \""
         << seqlookup[i->second->getLabel()] << "\\n\"" << endl;
    if(!i->second->getMark())
      printGraph(*(i->second));
    }
  }


/*Lookup function for a hash_map.  Returns 1 if -i- is in -Map-.
*/
bool lookup(const hash_map<unsigned int, Vertex>& Map, const unsigned i)
  {
  hash_map<unsigned int, Vertex>::const_iterator it = Map.find(i);
  return (it != Map.end());
  }



/* Get a hash value for a string.
   -- uses the algorithm from "art of programming"
*/
unsigned int DEKHash(const string& str)
  {
  unsigned int hash = static_cast<unsigned int>(str.length());
  for(size_t i = 0; i < str.length(); i++)
    hash = ((hash << 5) ^ (hash >> 27)) ^ str[i];
  seqlookup[hash] = str;  //Store the reverse lookup.
  return hash;
  }
