/* Krister Swenson                                                 Winter 2007-8

Functions for taking the maximum independent set of circle graphs.
*/

#include "miscg.h"

using namespace std;
using namespace __gnu_cxx;



/*Return in -mis- a maximum independent set of the given chord model.
The value returned is size of the set.
  ** this is from the paper "A New Simple Algorithm for the Maximum-Weight
     Independent Set Problem on Circle Graphs" by Valiente.
  ** It runs in O(dn) time where d is the maximum number of intervals
     at a particular index.
*/
int getMISCG(vector<int> &model, list<int> &mis)
  {
  int msize = (int)model.size();

    //First remap the labels in the chord model:
  hash_map<int,int> map;             //Map labels to [0,n-1]
  vector<int> invmap;                //Map [0,n-1] to labels.
  int remap=0;
  for(vector<int>::iterator i = model.begin(); i != model.end(); i++)
    {
    if(map.find(*i) != map.end())    //If we have already mapped this one
      *i = map[*i];                  //then keep it consistent.
    else                             //Otherwise,
      {
      map[*i] = remap;               //record the new label,
      invmap.push_back(*i);          //record the inverse map,
      *i = remap++;                  //and assign a new label.
      }
    }


    //Build the edges (the -all- list) and innitialize the data structures:
  MISlist *all = NULL;
  Tstr *T = new Tstr[msize+1];       //The scratch area for MIS on subproblems.
  int *righteps = new int[msize/2];  //The right endpoint of a chord.
  for(int i=0; i<msize/2; i++)
    righteps[i] = -1;
  for(int i=msize-1; i>=0; i--)      //For every endpoint,
    {
    if(righteps[model[i]] < 0)       //if it is the right endpoint
      righteps[model[i]] = i;        //then record it's position.
    else
      {                              //Otherwise, add an edge to the list.
      MISlist *tmp = new MISlist;
      tmp->next = all;
      tmp->e = new MISe;
      tmp->e->chord = model[i];
      tmp->e->weight = 1;
      tmp->e->left = i;
      tmp->e->right = righteps[model[i]];
      tmp->e->length = righteps[model[i]]-i;
      tmp->e->l = NULL;
      all = tmp;
      T[i].e = tmp->e;
      T[righteps[model[i]]].e = tmp->e;
      T[i].w = tmp->e->weight;
      T[righteps[model[i]]].w = tmp->e->weight;
      }
    }

    //Bin the edges by length (by the way -all- is built the ties will be in
    // right to left order):
  MISlist **buckets = new MISlist*[msize];
  for(int i=0; i<msize; i++)
    buckets[i] = NULL;
  MISlist *current = all;
  while(current != NULL)
    {
    MISlist *tmp = new MISlist;
    tmp->next = buckets[current->e->length];
    tmp->e = current->e;
    buckets[current->e->length] = tmp;
    current = current->next;
    }


    //Find the MIS for each interval on its own:
  for(int i=1; i<msize; i++) //Visit intervals nondecreasing, right to left:
    {
    current = buckets[i];
    while(current != NULL)
      {
      T[current->e->left].w = 1;
      T[current->e->left].nextt = NULL;
      T[current->e->right].w = 1;
      T[current->e->right].nextt = NULL;
      if(current->e->length > 2)
        {
        T[current->e->right].nextt = NULL;
        T[current->e->right].used = false;
        for(int k = current->e->right-1; k > current->e->left; k--)
          {
          T[k].w = T[k+1].w;                  //Weight if we don't choose this.
          int rind = T[k].e->right;
          if((rind != k) &&                   //If k is the left index
             (rind < current->e->right) &&    //and it is inside the current
             (T[k].w < T[k].e->weight+T[rind+1].w)) //and it is heavier,
            {                                 //we add it to the current.
            T[k].w = T[k].e->weight+T[rind+1].w;
            T[k].nextt = &T[rind+1];
            T[k].used = true;
            }
          else                                //Otherwise,
            {
            T[k].nextt = &T[k+1];             //mark that we use the one to
            T[k].used = false;                //the right.
            }
          }
        Tstr *tp = &T[current->e->left+1];    //We traverse the Ts with -tp-.
        while(tp != NULL)                     //Recover the subsolution:
          {
          if(tp->used)                        //If this subproblem was used,
            {
            MISlist *tmp = new MISlist;       //then add it to the current MIS.
            tmp->next = current->e->l;
            current->e->l = tmp;
            current->e->l->e = tp->e;
            current->e->weight += tp->w;
            }
          tp = tp->nextt;
          }
        }
      current = current->next;
      }
    }


    //Find the maximum independent set for the whole thing:
  T[msize-1].w = 0;
  T[msize-1].nextt = NULL;
  T[msize-1].used = false;
  T[msize].w = 0;
  T[msize].nextt = NULL;
  T[msize].used = false;
  for(int i=msize-2; i>=0; i--)
    {
    T[i].w = T[i+1].w;                       //Inherit the score from the right.
    int rind = T[i].e->right;
    if((rind != i) &&                        //If i is the right index
       (T[i].w < T[i].e->weight+T[rind+1].w))//and it contributes
      {
      T[i].w = T[i].e->weight+T[rind+1].w;   //then we take it on.
      T[i].nextt = &T[rind+1];
      T[i].used = true;
      }
    else                                     //Otherwise,
      {
      T[i].nextt = &T[i+1];                  //mark that we use the inherited
      T[i].used = false;                     //score from the right.
      }
    }

  Tstr *tp = &T[0];                          //We traverse the Ts wih -tp-.
  while(tp != NULL)                          //Recover the subsolution:
    {
    if(tp->used)                             //If this subproblem was used,
      {
      mis.push_back(tp->e->chord);           //then add the interval
      getSubMIS(tp->e->l, mis);              //and subintervals.
      }
    tp = tp->nextt;
    }


    //Remap the chord labels:
  for(list<int>::iterator i = mis.begin(); i != mis.end(); i++)
    *i = invmap[*i];


    //Free the memory:
  delete [] righteps;
  current = all;
  while(current != NULL)
    {
    MISlist *tmp = current;
    current = current->next;
    MISlist *tmp2 = tmp->e->l;
    while(tmp2 != NULL)
      {
      MISlist *tmp3 = tmp2->next;
      delete tmp2;
      tmp2 = tmp3;
      }
    delete tmp->e;
    delete tmp;
    }
  for(int i=0; i<msize; i++)
    {
    current = buckets[i];
    while(current != NULL)
      {
      MISlist *tmp = current;
      current = current->next;
      delete tmp;
      }
    }
  delete [] buckets;
  delete [] T;

  return mis.size();
  }

/*Print all the MISe that are part of this interval.
*/
void getSubMIS(MISlist *elist, list<int> &mis)
  {
  while(elist != NULL)
    {
    mis.push_back(elist->e->chord);
    getSubMIS(elist->e->l, mis);
    elist = elist->next;
    }
  }

int getMISCG(list<int> &model, list<int> &mis)
  {
  vector<int> v;
  std::copy(model.begin(), model.end(), back_inserter(v));
  return getMISCG(v, mis);
  }
