/* Krister Swenson                                                 Winter 2010
*
*  This is a trace kept in normal form.
*/


#include "Trace.h"
#include "Word.h"

using namespace std;


Trace::Trace()
  {
  normalform = list<Word>();
  }

/*Create a trace with one word.
*/
Trace::Trace(const Word &w)
  { normalform.push_back(w); } 

/*Create a normal form from the given word (list of sets).
*/
Trace::Trace(const list<Element> &invs)
  {
  for(list<Element>::const_iterator element = invs.begin();
      element != invs.end(); element++)
    addElement(*element);
  }

/*Create a normal form by appending the given word to the given trace.
 */
Trace::Trace(const Trace &t, const Word &w)
  {
  *this = t;
  appendWord(w);
  }

Trace::Trace(const Trace &t)
  { copy(t); }

void Trace::copy(const Trace &t)
  {
  normalform = list<Word>(t.normalform.size());
  std::copy(t.normalform.begin(), t.normalform.end(), normalform.begin());
  }

void Trace::operator=(const Trace &t)
  { normalform.clear(); copy(t); }

/*Append the given word to the Trace.
 */
void Trace::appendWord(const Word &w)
  { normalform.push_back(w); }


/*Remove the last element from the Trace.
 */
void Trace::removeLastElement()
  {
  if(normalform.empty())
    return;

  list<Word>::reverse_iterator lastword = normalform.rbegin();
  if(lastword->getSize() > 1)
    lastword->removeLast();
  else
    normalform.pop_back();
  }

/*Get the last element from the Trace.
 */
Element Trace::getLastElement() const
  {
  if(normalform.empty())
    return Element();

  return normalform.back().getLast();
  }


/*Remove the last word from the Trace.
 */
void Trace::removeLast()
  {
  if(normalform.empty())
    return;
  normalform.pop_back();
  }

/*Remove all but the first word.
 */
void Trace::removeAllButFirst()
  {
  for(int i=normalform.size(); i > 1; i--)
    normalform.pop_back();
  }




/*Add the given element to the normal form.
*/
void Trace::addElement(const Element &e)
  {                                        //If e doesn't commute with the last
  if(normalform.empty() || !normalform.back().commutesWith(e))
    {
    normalform.push_back(Word(e));         //word then add a new word.
    return;
    }

  list<Word>::reverse_iterator lastone = normalform.rbegin();
  list<Word>::reverse_iterator w = lastone;
  while(++w != normalform.rend())
    {
    if(!w->commutesWith(e))
      break;
    lastone = w;
    }

  lastone->addElement(e);
  }

/*Return true if the given element would be the last element of the trace if
 * it were added.
*/
bool Trace::goesAtEnd(const Element &e) const
  {
Element tmp;
tmp.insert(1);
tmp.insert(2);
tmp.insert(3);
tmp.insert(4);
tmp.insert(5);
tmp.insert(6);

  if(normalform.empty())
    return true;
  if(!normalform.back().commutesWith(e))   //If -e- belongs in a new word
    return true;

  list<Word>::const_reverse_iterator secondtolastword = normalform.rbegin();
  secondtolastword++;                      //Get the second to last word.
  if(secondtolastword == normalform.rend() ||
     !secondtolastword->commutesWith(e))   //If the element doesn't commute
    {                                      //with the second to last word
    string estring = getInvSetStr(e);      //then make its string.
    for(set<Element>::const_iterator i=normalform.back().getItBegin();
        i != normalform.back().getItEnd(); i++)
      {
      string e2string = getInvSetStr(*i);  //If -e- should go before any elems
      if(Word::comes_before(estring, e2string))
        return false;                      //of the last word then return false.
      }

    return true;                           //Otherwise, it belongs at the end
    }                                      //of the last word.

  return false;                            //It belongs in an earlier word.
  }


/*Return true if the given element belongs in the first word.
*/
bool Trace::goesInFirstWord(const Element &e) const
  {
  for(list<Word>::const_reverse_iterator w=normalform.rbegin();
      w!=normalform.rend(); w++)
    if(!w->commutesWith(e))
      return false;

  return true;
  }


/*Modify the normal form as though it were sorting from the identity to
 * the given permutation.
*/
void Trace::invert(const vector<int> &p)
  {
  reverse();             //Reverse the ordering on this trace.
                         //Now remap the elements:
  for(list<Word>::iterator w=normalform.begin(); w != normalform.end(); w++)
    w->remap(p);
  }

/*Rename the normal so that the sequence corresponds to sorting from the
 * identity to the given permutation.
*/
void Trace::rename(const vector<int> &p)
  {
                         //Now remap the elements:
  for(list<Word>::iterator w=normalform.begin(); w != normalform.end(); w++)
    w->remap(p);
  }


/*Reverse the order of the normal form.
*/
void Trace::reverse()
  {
  Trace t;
  
  for(list<Word>::const_reverse_iterator w=normalform.rbegin();
      w != normalform.rend(); w++)
    for(set<Element>::const_reverse_iterator e=w->getRevItBegin();
        e != w->getRevItEnd(); e++)
      t.addElement(*e);

  *this = t;
  }




string Trace::getString() const
  {
  stringstream output;
  if(normalform.empty())
    return "";

  for(list<Word>::const_iterator w = normalform.begin(); 
      w != --(normalform.end()); w++)
    output << *w << " < ";
  output << normalform.back();
  return output.str();
  }

ostream& operator<<(ostream& output, const Trace& t)
  {
  if(t.normalform.empty())
    return output;

  for(list<Word>::const_iterator w = t.normalform.begin(); 
      w != --(t.normalform.end()); w++)
    output << *w << " < ";
  output << t.normalform.back();

  return output;
  }
